%{

open List

open Ast_builder
open Syntax.Ast

let this = Id.mk_var "this"

%}

%token CLASS INTERFACE EXTENSION EXTENDS IMPLEMENTATION 
%token IMPLEMENTS STATIC NEW NULL IF ELSE 
%token WHILE TRUE FALSE WHERE EXISTS RETURN OF
%token ASSIGN EQ NEQ LEQ GEQ AND OR NOT PLUS MINUS 
%token TIMES DIV COMMA SEMI DOT AT
%token LBRACE RBRACE LSQBRACKET RSQBRACKET LABRACKET RABRACKET LPAREN RPAREN
%token <string> ID
%token <int> INT
%token EOF 

/* dummy tokens */
%token CAST

/* from lowest ... */
%right SEMI
%left OR
%left AND
%nonassoc EQ NEQ
%nonassoc LT GT LEQ GEQ
%left PLUS MINUS        
%left TIMES DIV
%nonassoc CAST
%right NEW
%right NOT 
%nonassoc UMINUS
%left DOT
/* ... to highest precedence */

%start cunit
%type <Syntax.Ast.definition list> cunit

%%

cunit: definitions EOF { $1 };

definitions: r_definitions { rev $1 };
r_definitions: /* empty */                 { [] }
             | r_definitions class_def     { Class_def $2 :: $1 }
             | r_definitions iface_def     { Iface_def $2 :: $1 }
             | r_definitions ext_def       { Ext_def $2 :: $1 }
             | r_definitions impl_def      { Impl_def $2 :: $1 }
             ;

class_def: CLASS id brack_tyvars extends constrs LBRACE class_body RBRACE
          { let (fields, methods) = Utils.split_either $7
             in mk_class $2 $3 $5 $4 fields methods };

extends: /* empty */       { None }
       | EXTENDS inst_type { Some (mk_inst_class $2) }
       ;

class_body: r_class_body { rev $1 };
r_class_body: /* empty */                 { [] }
            | r_class_body field_dec      { Utils.Left $2 :: $1 }
            | r_class_body method_def     { Utils.Right $2 :: $1 }
            ;

field_dec: ty id opt_assign SEMI  { { fdecl_type = $1;
                                      fdecl_name = Id.mk_field $2;
                                      fdecl_init = $3 } };

opt_assign: /* empty */    { None }
          | ASSIGN expr    { Some $2 }
          ;

/* to avoid shift/reduce conflict, we use a seemingly more complex grammer 
   than necessary */
method_def: method_def_rest                        { mk_mdef [] Nonstatic $1 }
          | brack_tyvars_ne method_def_rest        { mk_mdef $1 Nonstatic $2 }
          | STATIC method_def_rest                 { mk_mdef [] Static $2 }
          | brack_tyvars_ne STATIC method_def_rest { mk_mdef $1 Static $3 }
          ;

method_def_rest: ty id LPAREN arglist RPAREN constrs LBRACE block RBRACE
                   { ($2, $6, $1, $4, $8) } ;

method_defs: r_method_defs { rev $1 }
r_method_defs: /* empty */              { [] }
             | r_method_defs method_def { $2 :: $1 }
             ;

method_dec: brack_tyvars modifier ty id LPAREN arglist RPAREN constrs SEMI
          { mk_method_dec $4 $2 $1 $8 $3 $6 };

method_decs: r_method_decs { rev $1 }
r_method_decs: /* empty */               { [] }
             | r_method_decs method_dec  { $2 :: $1 }
             ;

modifier: /* empty */  { Nonstatic }
        | STATIC       { Static }
        ;

arglist: /* empty */  { [] }
       | ne_arglist   { rev $1 }
       ;

ne_arglist: arg                   { [$1] }
          | ne_arglist COMMA arg  { $3::$1 }
          ;

arg: ty id         { ($1, Id.mk_var $2) };

iface_def: INTERFACE id brack_tyvars LSQBRACKET tyvars RSQBRACKET constrs
             LBRACE iface_body RBRACE
         { let (smdecs, cmdecs) = Utils.split_either $9
            in mk_iface_def $5 $2 $3 $7 smdecs cmdecs };

iface_body: r_iface_body { rev $1 }
r_iface_body: /* empty */                           { [] }
            | r_iface_body method_dec               { Utils.Left $2 :: $1 }
            | r_iface_body CLASS id LBRACE method_decs RBRACE 
                { Utils.Right ($3,$5) :: $1 }
            ;

ext_def: EXTENSION id brack_tyvars constrs LBRACE method_defs RBRACE  
       { mk_ext_def $2 $3 $4 $6 };

impl_def: IMPLEMENTATION brack_tyvars tylist OF inst_type
            constrs LBRACE method_defs RBRACE
            { mk_impl_def $2 $5 $3 $6 $8 };

constrs: r_constrs            { rev $1 }
r_constrs: /* empty */        { [] }
         | r_constrs constr   { $2 :: $1 }
         ;

constr: WHERE tylist IMPLEMENTS impl_annot inst_type
          { mk_iface_constr $2 $4 $5 }
      | WHERE id EXTENDS inst_type                           
          { mk_class_constr $2 $4 }
      ;

impl_annot: /* empty */   { Bound_explicit }
         | NOT            { Bound_implicit }
         ;

inst_type: id tyargs       { ($1, $2) };

inst_type_ne: id tyargs_ne { Ty_class (mk_inst_class ($1, $2)) };

tyargs: /* empty */ { [] }
      | tyargs_ne   { rev $1 }
      ;
tyargs_ne: LABRACKET tyargs_aux RABRACKET  { rev $2 };
tyargs_aux: ty                   { [$1] }
          | tyargs_aux COMMA ty  { $3 :: $1 }
          ;

/* types need post-processing: 
   + X maybe a type variable or a class. 
*/
ty: ty_wo_parens      { $1 }
  | LPAREN ty RPAREN  { $2 }
  ;

/* in some situations, types cannot have parenthesis at the toplevel */
ty_wo_parens: inst_type  { let (t,ts) = $1 in  Ty_unknown (Id.mk_unknown t, ts) }
            | ex_ty      { $1 }
            ;

ex_ty: AT ty                      { Ty_exact $2 }
     | EXISTS ids constrs DOT ty  { mk_existential $2 $3 $5 }
     ;

tylist:   r_tylist          { rev $1 };
r_tylist: ty                { [$1] }
        | r_tylist COMMA ty { $3 :: $1 }
        ;

brack_tylist: /* empty */                 { [] }
            | LABRACKET tylist RABRACKET  { $2 }

block: r_block { rev $1 }
r_block: /* empty */       { [] }
       | r_block stmt      { $2 :: $1 }
       ;

stmt: IF LPAREN expr RPAREN LBRACE block RBRACE
        { Stmt_if($3, $6) }
    | IF LPAREN expr RPAREN LBRACE block RBRACE ELSE LBRACE block RBRACE
        { Stmt_if_else($3, $6, $10) }
    | WHILE LPAREN expr RPAREN LBRACE block RBRACE 
        { Stmt_while($3, $6) }
    | RETURN expr SEMI                 { Stmt_return $2 }
    | ty_wo_parens id opt_assign SEMI  { Stmt_vardecl ($1, Id.mk_var $2, $3) }
    | stmt_expr SEMI                   { Stmt_expr $1 }
    | expr DOT id ASSIGN expr SEMI     { Stmt_assign_field ($1, Id.mk_field $3, $5) }
    | id ASSIGN expr SEMI              { Stmt_assign_var (Id.mk_var $1, $3) }
    ;

expr: expr DOT id                        { Expr_field ($1, Id.mk_field $3) }
    | id                                 { Expr_var (Id.mk_var $1) }
    | INT                                { Expr_int $1 }
    | TRUE                               { Expr_bool true }
    | FALSE                              { Expr_bool false }
    | NULL                               { Expr_null }
    | stmt_expr                          { $1 }
    | LPAREN expr RPAREN                 { $2 }
    | cast_expr                          { $1 }
    ;

/* The precedence specification resolves a shift/reduce conflict.
   Without it, the string '(C) e1 . e2' could be either parsed as
   '((C) e1) . e2' (the parser reduces when looking at '.') or
   '(C) (e1.e2)' (the parser shifts when looking at '.'). The
   precedence specification below cause the 2nd parse to be 
   selected because DOT has a higher precedence than CAST. */
cast_expr: LPAREN expr RPAREN expr  %prec CAST        { mk_cast_from_expr $2 $4 }
         | LPAREN inst_type_ne RPAREN expr %prec CAST { Expr_cast ($2, $4) }
         | LPAREN ex_ty RPAREN expr %prec CAST        { Expr_cast ($2, $4) }
         ;

stmt_expr: expr dot_call                                   { mk_mcall $1 $2 }
         | inst_type LSQBRACKET tylist RSQBRACKET dot_call { mk_icall $1 $3 $5 }
         | NEW inst_type LPAREN RPAREN %prec NEW           { Expr_new 
                                                               (mk_inst_class $2) }
         ;

dot_call: DOT id brack_tylist LPAREN exprlist RPAREN { ($2, $3, $5) }

exprlist: /* empty */  { [] }
       | r_exprlist    { rev $1 }
       ;

r_exprlist: expr                   { [$1] }
          | r_exprlist COMMA expr  { $3::$1 }
          ;

ids: r_ids             { rev $1 };
r_ids: id              { [$1] }
     | r_ids COMMA id  { $3::$1 }
     ;

id: ID  { $1 };


tyvars: ids    { map Id.mk_tyvar $1 }

brack_tyvars: /* empty */      { [] }
            | brack_tyvars_ne  { $1 }

brack_tyvars_ne: LABRACKET tyvars RABRACKET  { ($2 : Id.tyvar list) }
